#include <tlhelp32.h>
#include <direct.h>
#include <string>
#include "common/IFileStream.h"
#include "loader_common/IdentifyEXE.h"
#include "loader_common/Error.h"
#include "Options.h"
#include "Inject.h"

IDebugLog	gLog("nvse_loader.log");

static bool InjectDLL(PROCESS_INFORMATION * info, const char * dllPath, ProcHookInfo * hookInfo);
static void PrintModuleInfo(UInt32 procID);

int main(int argc, char ** argv)
{
	gLog.SetPrintLevel(IDebugLog::kLevel_FatalError);
	gLog.SetLogLevel(IDebugLog::kLevel_DebugMessage);

	if(!g_options.Read(argc, argv))
	{
		PrintLoaderError("Couldn't read arguments.");
		g_options.PrintUsage();

		return -1;
	}

	if(g_options.m_optionsOnly)
	{
		g_options.PrintUsage();
		return 0;
	}

	if(g_options.m_verbose)
		gLog.SetPrintLevel(IDebugLog::kLevel_VerboseMessage);

	if(g_options.m_launchCS)
		_MESSAGE("launching editor");

	// create the process
	bool	dllHasFullPath = false;

	const char	* procName = g_options.m_launchCS ? "GECK.exe" : "FalloutNV.exe";
	const char	* baseDllName = g_options.m_launchCS ? "nvse_editor" : "nvse";

	char	currentWorkingDirectory[4096];
	ASSERT(_getcwd(currentWorkingDirectory, sizeof(currentWorkingDirectory)));

	std::string procPath = std::string(currentWorkingDirectory) + "\\" + std::string(procName);

	if(g_options.m_altEXE.size())
	{
		procPath = g_options.m_altEXE;
		_MESSAGE("launching alternate exe (%s)", procPath.c_str());
	}
	else if(!g_options.m_launchCS)
	{
		IFileStream	fileCheck;
		if(!fileCheck.Open(procPath.c_str()))
		{
#if 0
			// ### not sure what the ng version is called
			const char * kNoGoreEXEName = "falloutnvng.exe";

			_MESSAGE("couldn't find %s, trying %s", procName, kNoGoreEXEName);

			procName = kNoGoreEXEName;
			procPath = std::string(currentWorkingDirectory) + "\\" + std::string(procName);

#else
			PrintLoaderError("Couldn't find %s.", procName);
			return -1;
#endif
		}
	}

	_MESSAGE("launching: %s (%s)", procName, procPath.c_str());

	if(g_options.m_altDLL.size())
	{
		baseDllName = g_options.m_altDLL.c_str();
		_MESSAGE("launching alternate dll (%s)", baseDllName);

		dllHasFullPath = true;
	}

	std::string		dllSuffix;
	ProcHookInfo	procHookInfo;

	if(!IdentifyEXE(procPath.c_str(), g_options.m_launchCS, &dllSuffix, &procHookInfo))
	{
		_ERROR("unknown exe");
		return -1;
	}

	_MESSAGE("hook call addr = %08X", procHookInfo.hookCallAddr);
	_MESSAGE("load lib addr = %08X", procHookInfo.loadLibAddr);

	if(g_options.m_crcOnly)
		return 0;

	// build dll path
	std::string	dllPath;
	if(dllHasFullPath)
	{
		dllPath = baseDllName;
	}
	else
	{
		dllPath = std::string(currentWorkingDirectory) + "\\" + baseDllName + "_" + dllSuffix + ".dll";
	}

	_MESSAGE("dll = %s", dllPath.c_str());

	// check to make sure the dll exists
	{
		IFileStream	tempFile;

		if(!tempFile.Open(dllPath.c_str()))
		{
			PrintLoaderError("Couldn't find NVSE DLL (%s). Please make sure you have installed NVSE correctly and are running it from your Fallout folder.", dllPath.c_str());
			return -1;
		}
	}

	if(procHookInfo.procType == kProcType_Steam)
	{
		if(g_options.m_skipLauncher)
		{
			const char * kAppID = "22380";	// nogore app ID may be 22490

			SetEnvironmentVariable("SteamAppID", kAppID);
			SetEnvironmentVariable("SteamGameId", kAppID);
		}
	}

	STARTUPINFO			startupInfo = { 0 };
	PROCESS_INFORMATION	procInfo = { 0 };

	startupInfo.cb = sizeof(startupInfo);

	if(!CreateProcess(
		procPath.c_str(),
		NULL,	// no args
		NULL,	// default process security
		NULL,	// default thread security
		TRUE,	// don't inherit handles
		CREATE_SUSPENDED,
		NULL,	// no new environment
		NULL,	// no new cwd
		&startupInfo, &procInfo))
	{
		PrintLoaderError("Launching %s failed (%d).", procPath.c_str(), GetLastError());
		return -1;
	}

	bool	injectionSucceeded = false;

	switch(procHookInfo.procType)
	{
		case kProcType_Steam:
			injectionSucceeded = true;
			break;

		case kProcType_Normal:
			if(InjectDLL(&procInfo, dllPath.c_str(), &procHookInfo))
			{
				injectionSucceeded = true;

				if(g_options.m_launchCS && g_options.m_loadScriptErrorHook)
				{
					_MESSAGE("loading script error hook");

					//	this hooks 005C51ED with
					//	
					//	int PrintScriptErr(int * outErr, const char * fmt, va_list args)
					//	{
					//		char	buf[0x1300];
					//		
					//		if(fmt)
					//		{
					//			_vsprintf(buf, fmt, args);
					//			if(strstr(buf, "SCRIPTS: Script \'DialogueCompileAndRun\'"))
					//			{
					//				MessageBox(NULL, buf, NULL, MB_OK);
					//			}
					//		}
					//		
					//		return 0;	// should probably be the result of _vsprintf
					//	}
					//	
					//	need to contact the person who originally wrote this to see if we can integrate it

					if(!InjectDLL(&procInfo, "NVGeckExt.dll", &procHookInfo))
					{
						_ERROR("couldn't inject NVGeckExt.dll");
					}
				}
			}
			break;

		default:
			HALT("impossible");
	}

	if(!injectionSucceeded)
	{
		PrintLoaderError("Couldn't inject DLL.");

		_ERROR("terminating process");

		TerminateProcess(procInfo.hProcess, 0);
	}
	else
	{
		_MESSAGE("launching");

		ResumeThread(procInfo.hThread);

		if(g_options.m_moduleInfo)
		{
			Sleep(1000 * 3);	// wait 3 seconds

			PrintModuleInfo(procInfo.dwProcessId);
		}

		if(g_options.m_waitForClose)
			WaitForSingleObject(procInfo.hProcess, INFINITE);
	}

	CloseHandle(procInfo.hProcess);
	CloseHandle(procInfo.hThread);

	return 0;
}

static bool InjectDLL(PROCESS_INFORMATION * info, const char * dllPath, ProcHookInfo * hookInfo)
{
	bool	result = false;

	// wrap DLL injection in SEH, if it crashes print a message
	__try {
		result = DoInjectDLL(info, dllPath, hookInfo);
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		PrintLoaderError("DLL injection failed. In most cases, this is caused by an overly paranoid software firewall or antivirus package. Disabling either of these may solve the problem.");
		result = false;
	}

	return result;
}

static void PrintModuleInfo(UInt32 procID)
{
	HANDLE	snap = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, procID);
	if(snap != INVALID_HANDLE_VALUE)
	{
		MODULEENTRY32	module;

		module.dwSize = sizeof(module);

		if(Module32First(snap, &module))
		{
			do 
			{
				_MESSAGE("%08Xx%08X %08X %s %s", module.modBaseAddr, module.modBaseSize, module.hModule, module.szModule, module.szExePath);
			}
			while(Module32Next(snap, &module));
		}
		else
		{
			_ERROR("PrintModuleInfo: Module32First failed (%d)", GetLastError());
		}

		CloseHandle(snap);
	}
	else
	{
		_ERROR("PrintModuleInfo: CreateToolhelp32Snapshot failed (%d)", GetLastError());
	}
}
